# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
import json
import logging
import sys

import pika

from swh.core.statsd import statsd
from swh.scheduler import get_scheduler

logger = logging.getLogger(__name__)


def utcnow():
    return datetime.datetime.now(tz=datetime.timezone.utc)


def get_listener(broker_url, queue_name, scheduler_backend):
    connection = pika.BlockingConnection(pika.URLParameters(broker_url))
    channel = connection.channel()

    channel.queue_declare(queue=queue_name, durable=True)

    exchange = "celeryev"
    routing_key = "#"
    channel.queue_bind(queue=queue_name, exchange=exchange, routing_key=routing_key)

    channel.basic_qos(prefetch_count=1000)

    channel.basic_consume(
        queue=queue_name, on_message_callback=get_on_message(scheduler_backend),
    )

    return channel


def get_on_message(scheduler_backend):
    def on_message(channel, method_frame, properties, body):
        try:
            events = json.loads(body)
        except Exception:
            logger.warning("Could not parse body %r", body)
            events = []

        if not isinstance(events, list):
            events = [events]

        for event in events:
            logger.debug("Received event %r", event)
            process_event(event, scheduler_backend)

        channel.basic_ack(delivery_tag=method_frame.delivery_tag)

    return on_message


def process_event(event, scheduler_backend):
    uuid = event.get("uuid")
    if not uuid:
        return

    event_type = event["type"]
    statsd.increment(
        "swh_scheduler_listener_handled_event_total", tags={"event_type": event_type}
    )

    if event_type == "task-started":
        scheduler_backend.start_task_run(
            uuid, timestamp=utcnow(), metadata={"worker": event.get("hostname")},
        )
    elif event_type == "task-result":
        result = event["result"]

        status = None

        if isinstance(result, dict) and "status" in result:
            status = result["status"]
            if status == "success":
                status = "eventful" if result.get("eventful") else "uneventful"

        if status is None:
            status = "eventful" if result else "uneventful"

        scheduler_backend.end_task_run(
            uuid, timestamp=utcnow(), status=status, result=result
        )
    elif event_type == "task-failed":
        scheduler_backend.end_task_run(uuid, timestamp=utcnow(), status="failed")


if __name__ == "__main__":
    url = sys.argv[1]
    logging.basicConfig(level=logging.DEBUG)
    scheduler_backend = get_scheduler("local", args={"db": "service=swh-scheduler"})
    channel = get_listener(url, "celeryev.test", scheduler_backend)
    logger.info("Start consuming")
    channel.start_consuming()



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import traceback

from click.testing import CliRunner
import pkg_resources
import pytest
import yaml

from swh.scheduler import get_scheduler
from swh.scheduler.cli import cli

FAKE_MODULE_ENTRY_POINTS = {
    "lister.gnu=swh.lister.gnu:register",
    "lister.pypi=swh.lister.pypi:register",
}


@pytest.fixture
def mock_pkg_resources(monkeypatch):
    """Monkey patch swh.scheduler's mock_pkg_resources.iter_entry_point call

    """

    def fake_iter_entry_points(*args, **kwargs):
        """Substitute fake function to return a fixed set of entrypoints

        """
        from pkg_resources import Distribution, EntryPoint

        d = Distribution()
        return [EntryPoint.parse(entry, dist=d) for entry in FAKE_MODULE_ENTRY_POINTS]

    original_method = pkg_resources.iter_entry_points
    monkeypatch.setattr(pkg_resources, "iter_entry_points", fake_iter_entry_points)

    yield

    # reset monkeypatch: is that needed?
    monkeypatch.setattr(pkg_resources, "iter_entry_points", original_method)


@pytest.fixture
def local_sched_config(swh_scheduler_config):
    """Expose the local scheduler configuration

    """
    return {"scheduler": {"cls": "local", **swh_scheduler_config}}


@pytest.fixture
def local_sched_configfile(local_sched_config, tmp_path):
    """Write in temporary location the local scheduler configuration

    """
    configfile = tmp_path / "config.yml"
    configfile.write_text(yaml.dump(local_sched_config))
    return configfile.as_posix()


def test_register_ttypes_all(
    mock_pkg_resources, local_sched_config, local_sched_configfile
):
    """Registering all task types"""

    for command in [
        ["--config-file", local_sched_configfile, "task-type", "register"],
        ["--config-file", local_sched_configfile, "task-type", "register", "-p", "all"],
        [
            "--config-file",
            local_sched_configfile,
            "task-type",
            "register",
            "-p",
            "lister.gnu",
            "-p",
            "lister.pypi",
        ],
    ]:
        result = CliRunner().invoke(cli, command)

        assert result.exit_code == 0, traceback.print_exception(*result.exc_info)

        scheduler = get_scheduler(**local_sched_config["scheduler"])
        all_tasks = [
            "list-gnu-full",
            "list-pypi",
        ]
        for task in all_tasks:
            task_type_desc = scheduler.get_task_type(task)
            assert task_type_desc
            assert task_type_desc["type"] == task
            assert task_type_desc["backoff_factor"] == 1


def test_register_ttypes_filter(
    mock_pkg_resources, local_sched_config, local_sched_configfile
):
    """Filtering on one worker should only register its associated task type

    """
    result = CliRunner().invoke(
        cli,
        [
            "--config-file",
            local_sched_configfile,
            "task-type",
            "register",
            "--plugins",
            "lister.gnu",
        ],
    )

    assert result.exit_code == 0, traceback.print_exception(*result.exc_info)

    scheduler = get_scheduler(**local_sched_config["scheduler"])
    all_tasks = [
        "list-gnu-full",
    ]
    for task in all_tasks:
        task_type_desc = scheduler.get_task_type(task)
        assert task_type_desc
        assert task_type_desc["type"] == task
        assert task_type_desc["backoff_factor"] == 1



# Copyright (C) 2016-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

# WARNING: do not import unnecessary things here to keep cli startup time under
# control
import logging
import time

import click

from . import cli


@cli.command("start-runner")
@click.option(
    "--period",
    "-p",
    default=0,
    help=(
        "Period (in s) at witch pending tasks are checked and "
        "executed. Set to 0 (default) for a one shot."
    ),
)
@click.pass_context
def runner(ctx, period):
    """Starts a swh-scheduler runner service.

    This process is responsible for checking for ready-to-run tasks and
    schedule them."""
    from swh.scheduler.celery_backend.config import build_app
    from swh.scheduler.celery_backend.runner import run_ready_tasks

    app = build_app(ctx.obj["config"].get("celery"))
    app.set_current()

    logger = logging.getLogger(__name__ + ".runner")
    scheduler = ctx.obj["scheduler"]
    logger.debug("Scheduler %s" % scheduler)
    try:
        while True:
            logger.debug("Run ready tasks")
            try:
                ntasks = len(run_ready_tasks(scheduler, app))
                if ntasks:
                    logger.info("Scheduled %s tasks", ntasks)
            except Exception:
                logger.exception("Unexpected error in run_ready_tasks()")
            if not period:
                break
            time.sleep(period)
    except KeyboardInterrupt:
        ctx.exit(0)


@cli.command("start-listener")
@click.pass_context
def listener(ctx):
    """Starts a swh-scheduler listener service.

    This service is responsible for listening at task lifecycle events and
    handle their workflow status in the database."""
    scheduler_backend = ctx.obj["scheduler"]
    if not scheduler_backend:
        raise ValueError("Scheduler class (local/remote) must be instantiated")

    broker = (
        ctx.obj["config"]
        .get("celery", {})
        .get("task_broker", "amqp://guest@localhost/%2f")
    )

    from swh.scheduler.celery_backend.pika_listener import get_listener

    listener = get_listener(broker, "celeryev.listener", scheduler_backend)
    try:
        listener.start_consuming()
    finally:
        listener.stop_consuming()


@cli.command("rpc-serve")
@click.option("--host", default="0.0.0.0", help="Host to run the scheduler server api")
@click.option("--port", default=5008, type=click.INT, help="Binding port of the server")
@click.option(
    "--debug/--nodebug",
    default=None,
    help=(
        "Indicates if the server should run in debug mode. "
        "Defaults to True if log-level is DEBUG, False otherwise."
    ),
)
@click.pass_context
def rpc_server(ctx, host, port, debug):
    """Starts a swh-scheduler API HTTP server.
    """
    if ctx.obj["config"]["scheduler"]["cls"] == "remote":
        click.echo(
            "The API server can only be started with a 'local' " "configuration",
            err=True,
        )
        ctx.exit(1)

    from swh.scheduler.api import server

    server.app.config.update(ctx.obj["config"])
    if debug is None:
        debug = ctx.obj["log_level"] <= logging.DEBUG
    server.app.run(host, port=port, debug=bool(debug))



# Copyright (C) 2019-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime

import elasticsearch
import pytest

from swh.scheduler.backend_es import get_elasticsearch

from ..common import TEMPLATES, tasks_from_template


def test_get_elasticsearch():
    with pytest.raises(ValueError, match="Unknown elasticsearch class"):
        get_elasticsearch("unknown")

    es = get_elasticsearch("memory")
    assert es
    from swh.scheduler.elasticsearch_memory import MemoryElasticsearch

    assert isinstance(es, MemoryElasticsearch)

    es = get_elasticsearch("local")
    assert es
    assert isinstance(es, elasticsearch.Elasticsearch)


def test_backend_setup_basic(swh_elasticsearch_backend):
    """Elastic search instance should allow to create/close/check index

    """
    index_name = "swh-tasks-2010-01"
    try:
        swh_elasticsearch_backend.storage.indices.get_mapping(index_name)
    except (elasticsearch.exceptions.NotFoundError, KeyError):
        pass

    assert not swh_elasticsearch_backend.storage.indices.exists(index_name)
    swh_elasticsearch_backend.create(index_name)
    assert swh_elasticsearch_backend.storage.indices.exists(index_name)
    assert swh_elasticsearch_backend.is_index_opened(index_name)

    # index exists with a mapping
    mapping = swh_elasticsearch_backend.storage.indices.get_mapping(index_name)
    assert mapping != {}


def test_backend_setup_index(swh_elasticsearch_backend):
    """Elastic search instance should allow to bulk index

    """
    template_git = TEMPLATES["git"]
    next_run_date = datetime.datetime.utcnow() - datetime.timedelta(days=1)
    tasks = tasks_from_template(template_git, next_run_date, 1)
    index_name = swh_elasticsearch_backend.compute_index_name(
        next_run_date.year, next_run_date.month
    )
    assert not swh_elasticsearch_backend.storage.indices.exists(index_name)

    tasks = list(swh_elasticsearch_backend.streaming_bulk(index_name, tasks))
    assert len(tasks) > 0

    for output_task in tasks:
        assert output_task is not None
        assert output_task["type"] == template_git["type"]
        assert output_task["arguments"] is not None
        next_run = output_task["next_run"]
        if isinstance(next_run, str):  # real elasticsearch
            assert next_run == next_run_date.isoformat()
        else:  # memory implem. does not really index
            assert next_run == next_run_date

    assert swh_elasticsearch_backend.storage.indices.exists(index_name)
    assert swh_elasticsearch_backend.is_index_opened(index_name)
    mapping = swh_elasticsearch_backend.storage.indices.get_mapping(index_name)
    assert mapping != {}



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

# WARNING: do not import unnecessary things here to keep cli startup time under
# control

import click

TASK_BATCH_SIZE = 1000  # Number of tasks per query to the scheduler


def schedule_origin_batches(scheduler, task_type, origins, origin_batch_size, kwargs):
    from itertools import islice

    from swh.scheduler.utils import create_task_dict

    nb_origins = 0
    nb_tasks = 0

    while True:
        task_batch = []
        for _ in range(TASK_BATCH_SIZE):
            # Group origins
            origin_batch = []
            for origin in islice(origins, origin_batch_size):
                origin_batch.append(origin)
            nb_origins += len(origin_batch)
            if not origin_batch:
                break

            # Create a task for these origins
            args = [origin_batch]
            task_dict = create_task_dict(task_type, "oneshot", *args, **kwargs)
            task_batch.append(task_dict)

        # Schedule a batch of tasks
        if not task_batch:
            break
        nb_tasks += len(task_batch)
        if scheduler:
            scheduler.create_tasks(task_batch)
        click.echo("Scheduled %d tasks (%d origins)." % (nb_tasks, nb_origins))

    # Print final status.
    if nb_tasks:
        click.echo("Done.")
    else:
        click.echo("Nothing to do (no origin metadata matched the criteria).")


def parse_argument(option):
    import yaml

    try:
        return yaml.safe_load(option)
    except Exception:
        raise click.ClickException("Invalid argument: {}".format(option))


def parse_options(options):
    """Parses options from a CLI as YAML and turns it into Python
    args and kwargs.

    >>> parse_options([])
    ([], {})
    >>> parse_options(['foo', 'bar'])
    (['foo', 'bar'], {})
    >>> parse_options(['[foo, bar]'])
    ([['foo', 'bar']], {})
    >>> parse_options(['"foo"', '"bar"'])
    (['foo', 'bar'], {})
    >>> parse_options(['foo="bar"'])
    ([], {'foo': 'bar'})
    >>> parse_options(['"foo"', 'bar="baz"'])
    (['foo'], {'bar': 'baz'})
    >>> parse_options(['42', 'bar=False'])
    ([42], {'bar': False})
    >>> parse_options(['42', 'bar=false'])
    ([42], {'bar': False})
    >>> parse_options(['42', '"foo'])
    Traceback (most recent call last):
      ...
    click.exceptions.ClickException: Invalid argument: "foo
    """
    kw_pairs = [x.split("=", 1) for x in options if "=" in x]
    args = [parse_argument(x) for x in options if "=" not in x]
    kw = {k: parse_argument(v) for (k, v) in kw_pairs}
    return (args, kw)



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
import hashlib
import logging
import random
from typing import Any, Dict

import pytest

from swh.scheduler.elasticsearch_memory import BasicSerializer, BasicTransport

from ..common import TEMPLATES, tasks_from_template

logger = logging.getLogger(__name__)


def test_serializer():
    s = BasicSerializer()
    assert s

    data = {"something": [1, 2, 3], "cool": {"1": "2"}}
    actual_data = s.dumps(data)

    assert isinstance(actual_data, str)
    assert actual_data == str(data)


def test_basic_transport():
    b = BasicTransport()
    assert b

    assert isinstance(b.serializer, BasicSerializer)


def test_index_manipulation(swh_elasticsearch_memory):
    index_name = "swh-tasks-xxxx"
    indices = swh_elasticsearch_memory.index

    assert not swh_elasticsearch_memory.exists(index_name)
    assert index_name not in indices

    # so stat raises
    with pytest.raises(Exception):
        swh_elasticsearch_memory.stats(index_name)

    # we create the index
    swh_elasticsearch_memory.create(index_name)

    # now the index exists
    assert swh_elasticsearch_memory.exists(index_name)
    assert index_name in indices
    # it's opened
    assert indices[index_name]["status"] == "opened"

    # so stats is happy
    swh_elasticsearch_memory.stats(index_name)

    # open the index, nothing changes
    swh_elasticsearch_memory.open(index_name)
    assert indices[index_name]["status"] == "opened"

    # close the index
    swh_elasticsearch_memory.close(index_name)

    assert indices[index_name]["status"] == "closed"

    # reopen the index (fun times)
    swh_elasticsearch_memory.open(index_name)
    assert indices[index_name]["status"] == "opened"


def test_bulk_and_mget(swh_elasticsearch_memory):
    # initialize tasks
    template_git = TEMPLATES["git"]
    next_run_start = datetime.datetime.utcnow() - datetime.timedelta(days=1)

    tasks = tasks_from_template(template_git, next_run_start, 100)

    def compute_id(stask):
        return hashlib.sha1(stask.encode("utf-8")).hexdigest()

    body = []
    ids_to_task = {}
    for task in tasks:
        date = task["next_run"]
        index_name = f"swh-tasks-{date.year}-{date.month}"
        idx = {"index": {"_index": index_name}}
        sidx = swh_elasticsearch_memory.transport.serializer.dumps(idx)
        body.append(sidx)

        stask = swh_elasticsearch_memory.transport.serializer.dumps(task)
        body.append(stask)

        _id = compute_id(stask)
        ids_to_task[_id] = task
        logger.debug(f"_id: {_id}, task: {task}")

    # store

    # create the index first
    swh_elasticsearch_memory.create(index_name)

    # then bulk insert new data
    result = swh_elasticsearch_memory.bulk("\n".join(body))

    # no guarantee in the order
    assert result
    actual_items = result["items"]
    assert len(actual_items) == len(ids_to_task)

    def get_id(data: Dict[str, Any]) -> str:
        return data["index"]["_id"]

    actual_items = sorted(actual_items, key=get_id)

    expected_items = {
        "items": [{"index": {"status": 200, "_id": _id}} for _id in list(ids_to_task)]
    }

    expected_items = sorted(expected_items["items"], key=get_id)
    assert actual_items == expected_items

    # retrieve

    nb_docs = 10
    ids = list(ids_to_task)
    random_ids = []
    # add some inexistent ids
    for i in range(16):
        noisy_id = f"{i}" * 40
        random_ids.append(noisy_id)
    random_ids.extend(random.sample(ids, nb_docs))  # add relevant ids
    for i in range(16, 32):
        noisy_id = f"{i}" * 40
        random_ids.append(noisy_id)

    result = swh_elasticsearch_memory.mget(index=index_name, body={"ids": random_ids})
    assert result["docs"]
    assert len(result["docs"]) == nb_docs, "no random and inexistent id found"
    for doc in result["docs"]:
        assert doc["found"]

        actual_task = doc["_source"]
        _id = compute_id(str(actual_task))
        expected_task = ids_to_task[_id]
        assert actual_task == expected_task



# Copyright (C) 2018-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


from swh.core.api import RPCClient

from .. import exc
from ..interface import SchedulerInterface
from .serializers import DECODERS, ENCODERS


class RemoteScheduler(RPCClient):
    """Proxy to a remote scheduler API

    """

    backend_class = SchedulerInterface

    reraise_exceptions = [getattr(exc, a) for a in exc.__all__]

    extra_type_decoders = DECODERS
    extra_type_encoders = ENCODERS



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging

from click.testing import CliRunner
import pytest

from swh.scheduler.cli import cli


def invoke(*args, catch_exceptions=False):
    result = CliRunner(mix_stderr=False).invoke(
        cli, ["celery-monitor", *args], catch_exceptions=catch_exceptions,
    )

    return result


def test_celery_monitor():
    """Check that celery-monitor returns its help text"""

    result = invoke()

    assert "Commands:" in result.stdout
    assert "Options:" in result.stdout


def test_celery_monitor_ping(
    caplog, swh_scheduler_celery_app, swh_scheduler_celery_worker
):
    caplog.set_level(logging.INFO, "swh.scheduler.cli.celery_monitor")

    result = invoke("--pattern", swh_scheduler_celery_worker.hostname, "ping-workers")

    assert result.exit_code == 0

    assert len(caplog.records) == 1

    (record,) = caplog.records

    assert record.levelname == "INFO"
    assert f"response from {swh_scheduler_celery_worker.hostname}" in record.message


@pytest.mark.parametrize(
    "filter_args,filter_message,exit_code",
    [
        ((), "Matching all workers", 0),
        (
            ("--pattern", "celery@*.test-host"),
            "Using glob pattern celery@*.test-host",
            1,
        ),
        (
            ("--pattern", "celery@test-type.test-host"),
            "Using destinations celery@test-type.test-host",
            1,
        ),
        (
            ("--pattern", "celery@test-type.test-host,celery@test-type2.test-host"),
            (
                "Using destinations "
                "celery@test-type.test-host, celery@test-type2.test-host"
            ),
            1,
        ),
    ],
)
def test_celery_monitor_ping_filter(
    caplog,
    swh_scheduler_celery_app,
    swh_scheduler_celery_worker,
    filter_args,
    filter_message,
    exit_code,
):
    caplog.set_level(logging.DEBUG, "swh.scheduler.cli.celery_monitor")

    result = invoke("--timeout", "1.5", *filter_args, "ping-workers")

    assert result.exit_code == exit_code, result.stdout

    got_no_response_message = False
    got_filter_message = False

    for record in caplog.records:
        # Check the proper filter has been generated
        if record.levelname == "DEBUG":
            if filter_message in record.message:
                got_filter_message = True
        # Check that no worker responded
        if record.levelname == "INFO":
            if "No response in" in record.message:
                got_no_response_message = True

    assert got_filter_message

    if filter_args:
        assert got_no_response_message


def test_celery_monitor_list_running(
    caplog, swh_scheduler_celery_app, swh_scheduler_celery_worker
):
    caplog.set_level(logging.DEBUG, "swh.scheduler.cli.celery_monitor")

    result = invoke("--pattern", swh_scheduler_celery_worker.hostname, "list-running")

    assert result.exit_code == 0, result.stdout

    for record in caplog.records:
        if record.levelname != "INFO":
            continue
        assert (
            f"{swh_scheduler_celery_worker.hostname}: no active tasks" in record.message
        )


@pytest.mark.parametrize("format", ["csv", "pretty"])
def test_celery_monitor_list_running_format(
    caplog, swh_scheduler_celery_app, swh_scheduler_celery_worker, format
):
    caplog.set_level(logging.DEBUG, "swh.scheduler.cli.celery_monitor")

    result = invoke(
        "--pattern",
        swh_scheduler_celery_worker.hostname,
        "list-running",
        "--format",
        format,
    )

    assert result.exit_code == 0, result.stdout

    for record in caplog.records:
        if record.levelname != "INFO":
            continue
        assert (
            f"{swh_scheduler_celery_worker.hostname}: no active tasks" in record.message
        )

    if format == "csv":
        lines = result.stdout.splitlines()
        assert lines == ["worker,name,args,kwargs,duration,worker_pid"]



# Copyright (C) 2016-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from importlib import import_module
import logging
from typing import Mapping

# WARNING: do not import unnecessary things here to keep cli startup time under
# control
import click
from pkg_resources import iter_entry_points

from . import cli

logger = logging.getLogger(__name__)


DEFAULT_TASK_TYPE = {
    "full": {  # for tasks like 'list_xxx_full()'
        "default_interval": "90 days",
        "min_interval": "90 days",
        "max_interval": "90 days",
        "backoff_factor": 1,
    },
    "*": {  # value if not suffix matches
        "default_interval": "1 day",
        "min_interval": "1 day",
        "max_interval": "1 day",
        "backoff_factor": 1,
    },
}


PLUGIN_WORKER_DESCRIPTIONS = {
    entry_point.name: entry_point for entry_point in iter_entry_points("swh.workers")
}


@cli.group("task-type")
@click.pass_context
def task_type(ctx):
    """Manipulate task types."""
    pass


@task_type.command("list")
@click.option("--verbose", "-v", is_flag=True, default=False, help="Verbose mode")
@click.option(
    "--task_type",
    "-t",
    multiple=True,
    default=None,
    help="List task types of given type",
)
@click.option(
    "--task_name",
    "-n",
    multiple=True,
    default=None,
    help="List task types of given backend task name",
)
@click.pass_context
def list_task_types(ctx, verbose, task_type, task_name):
    click.echo("Known task types:")
    if verbose:
        tmpl = (
            click.style("{type}: ", bold=True)
            + """{backend_name}
  {description}
  interval: {default_interval} [{min_interval}, {max_interval}]
  backoff_factor: {backoff_factor}
  max_queue_length: {max_queue_length}
  num_retries: {num_retries}
  retry_delay: {retry_delay}
"""
        )
    else:
        tmpl = "{type}:\n  {description}"
    for tasktype in sorted(
        ctx.obj["scheduler"].get_task_types(), key=lambda x: x["type"]
    ):
        if task_type and tasktype["type"] not in task_type:
            continue
        if task_name and tasktype["backend_name"] not in task_name:
            continue
        click.echo(tmpl.format(**tasktype))


@task_type.command("register")
@click.option(
    "--plugins",
    "-p",
    "plugins",
    multiple=True,
    default=("all",),
    type=click.Choice(["all"] + list(PLUGIN_WORKER_DESCRIPTIONS)),
    help="Registers task-types for provided plugins. " "Defaults to all",
)
@click.pass_context
def register_task_types(ctx, plugins):
    """Register missing task-type entries in the scheduler.

    According to declared tasks in each loaded worker (e.g. lister, loader,
    ...) plugins.

    """
    import celery.app.task

    scheduler = ctx.obj["scheduler"]

    if plugins == ("all",):
        plugins = list(PLUGIN_WORKER_DESCRIPTIONS)

    for plugin in plugins:
        entrypoint = PLUGIN_WORKER_DESCRIPTIONS[plugin]
        logger.info("Loading entrypoint for plugin %s", plugin)
        registry_entry = entrypoint.load()()

        for task_module in registry_entry["task_modules"]:
            mod = import_module(task_module)
            for task_name in (x for x in dir(mod) if not x.startswith("_")):
                logger.debug("Loading task name %s", task_name)
                taskobj = getattr(mod, task_name)
                if isinstance(taskobj, celery.app.task.Task):
                    tt_name = task_name.replace("_", "-")
                    task_cfg = registry_entry.get("task_types", {}).get(tt_name, {})
                    ensure_task_type(task_module, tt_name, taskobj, task_cfg, scheduler)


def ensure_task_type(
    task_module: str, task_type: str, swhtask, task_config: Mapping, scheduler
):
    """Ensure a given task-type (for the task_module) exists in the scheduler.

    Args:
        task_module: task module we are currently checking for task type
            consistency
        task_type: the type of the task to check/insert (correspond to
            the 'type' field in the db)
        swhtask (SWHTask): the SWHTask instance the task-type correspond to
        task_config: a dict with specific/overloaded values for the
            task-type to be created
        scheduler: the scheduler object used to access the scheduler db

    """
    for suffix, defaults in DEFAULT_TASK_TYPE.items():
        if task_type.endswith("-" + suffix):
            task_type_dict = defaults.copy()
            break
    else:
        task_type_dict = DEFAULT_TASK_TYPE["*"].copy()

    task_type_dict["type"] = task_type
    task_type_dict["backend_name"] = swhtask.name
    if swhtask.__doc__:
        task_type_dict["description"] = swhtask.__doc__.splitlines()[0]

    task_type_dict.update(task_config)

    current_task_type = scheduler.get_task_type(task_type)
    if current_task_type:
        # Ensure the existing task_type is consistent in the scheduler
        if current_task_type["backend_name"] != task_type_dict["backend_name"]:
            logger.warning(
                "Existing task type %s for module %s has a "
                "different backend name than current "
                "code version provides (%s vs. %s)",
                task_type,
                task_module,
                current_task_type["backend_name"],
                task_type_dict["backend_name"],
            )
    else:
        logger.info("Create task type %s in scheduler", task_type)
        logger.debug("  %s", task_type_dict)
        scheduler.create_task_type(task_type_dict)


@task_type.command("add")
@click.argument("type", required=True)
@click.argument("task-name", required=True)
@click.argument("description", required=True)
@click.option(
    "--default-interval",
    "-i",
    default="90 days",
    help='Default interval ("90 days" by default)',
)
@click.option(
    "--min-interval",
    default=None,
    help="Minimum interval (default interval if not set)",
)
@click.option(
    "--max-interval",
    "-i",
    default=None,
    help="Maximal interval (default interval if not set)",
)
@click.option("--backoff-factor", "-f", type=float, default=1, help="Backoff factor")
@click.pass_context
def add_task_type(
    ctx,
    type,
    task_name,
    description,
    default_interval,
    min_interval,
    max_interval,
    backoff_factor,
):
    """Create a new task type
    """
    scheduler = ctx.obj["scheduler"]
    if not scheduler:
        raise ValueError("Scheduler class (local/remote) must be instantiated")
    task_type = dict(
        type=type,
        backend_name=task_name,
        description=description,
        default_interval=default_interval,
        min_interval=min_interval,
        max_interval=max_interval,
        backoff_factor=backoff_factor,
        max_queue_length=None,
        num_retries=None,
        retry_delay=None,
    )
    scheduler.create_task_type(task_type)
    click.echo("OK")



# Copyright (C) 2016-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from datetime import datetime, timezone
import os
from typing import List

import pytest

from swh.scheduler.model import ListedOrigin, Lister
from swh.scheduler.tests.common import LISTERS

# make sure we are not fooled by CELERY_ config environment vars
for var in [x for x in os.environ.keys() if x.startswith("CELERY")]:
    os.environ.pop(var)


# test_cli tests depends on a en/C locale, so ensure it
os.environ["LC_ALL"] = "C.UTF-8"


@pytest.fixture
def stored_lister(swh_scheduler) -> Lister:
    """Store a lister in the scheduler and return its information"""
    return swh_scheduler.get_or_create_lister(**LISTERS[0])


@pytest.fixture
def listed_origins(stored_lister) -> List[ListedOrigin]:
    """Return a (fixed) set of 1000 listed origins"""
    return [
        ListedOrigin(
            lister_id=stored_lister.id,
            url=f"https://example.com/{i:04d}.git",
            visit_type="git",
            last_update=datetime(2020, 6, 15, 16, 0, 0, i, tzinfo=timezone.utc),
        )
        for i in range(1000)
    ]



# Copyright (C) 2015-2019 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from datetime import datetime

from celery import current_app
import celery.app.task
from celery.utils.log import get_task_logger

from swh.core.statsd import Statsd


def ts():
    return int(datetime.utcnow().timestamp())


class SWHTask(celery.app.task.Task):
    """a schedulable task (abstract class)

    Current implementation is based on Celery. See
    http://docs.celeryproject.org/en/latest/reference/celery.app.task.html for
    how to use tasks once instantiated

    """

    _statsd = None
    _log = None

    @property
    def statsd(self):
        if self._statsd:
            return self._statsd
        worker_name = current_app.conf.get("worker_name")
        if worker_name:
            self._statsd = Statsd(
                constant_tags={"task": self.name, "worker": worker_name,}
            )
            return self._statsd
        else:
            statsd = Statsd(
                constant_tags={"task": self.name, "worker": "unknown worker",}
            )
            return statsd

    def __call__(self, *args, **kwargs):
        self.statsd.increment("swh_task_called_count")
        self.statsd.gauge("swh_task_start_ts", ts())
        with self.statsd.timed("swh_task_duration_seconds"):
            result = super().__call__(*args, **kwargs)
            try:
                status = result["status"]
                if status == "success":
                    status = "eventful" if result.get("eventful") else "uneventful"
            except Exception:
                status = "eventful" if result else "uneventful"

            self.statsd.gauge("swh_task_end_ts", ts(), tags={"status": status})
            return result

    def on_failure(self, exc, task_id, args, kwargs, einfo):
        self.statsd.increment("swh_task_failure_count")

    def on_success(self, retval, task_id, args, kwargs):
        self.statsd.increment("swh_task_success_count")
        # this is a swh specific event. Used to attach the retval to the
        # task_run
        self.send_event("task-result", result=retval)

    @property
    def log(self):
        if self._log is None:
            self._log = get_task_logger(self.name)
        return self._log

    def run(self, *args, **kwargs):
        self.log.debug("%s: args=%s, kwargs=%s", self.name, args, kwargs)
        ret = super().run(*args, **kwargs)
        self.log.debug("%s: OK => %s", self.name, ret)
        return ret
